# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for vizier.pythia.policies.random_policy."""
from vizier import pythia
from vizier import pyvizier
from vizier._src.algorithms.policies import random_policy
from vizier.testing import test_studies

from absl.testing import absltest


class RandomPolicyTest(absltest.TestCase):
  # TODO: Add conditional test case.

  def setUp(self):
    """Setups up search space and policy."""
    self.study_config = pyvizier.ProblemStatement(
        test_studies.flat_space_with_all_types()
    )
    self.policy_supporter = pythia.InRamPolicySupporter(self.study_config)
    self.policy = random_policy.RandomPolicy(
        policy_supporter=self.policy_supporter
    )
    super().setUp()

  def test_make_suggestions(self):
    """Tests random parameter generation wrapped around Policy."""
    num_suggestions = 5
    num_params = len(self.study_config.search_space.parameters)

    suggest_request = pythia.SuggestRequest(
        study_descriptor=self.policy_supporter.study_descriptor(),
        count=num_suggestions,
    )
    decisions = self.policy.suggest(suggest_request)
    self.assertLen(decisions.suggestions, num_suggestions)
    self.assertFalse(decisions.metadata)
    for suggestion in decisions.suggestions:
      self.assertLen(suggestion.parameters, num_params)

  def test_make_early_stopping_decisions(self):
    """Checks if all ACTIVE/PENDING trials become completed in random order."""
    count = 10
    _ = self.policy_supporter.SuggestTrials(self.policy, count=count)

    request_trial_ids = [1, 2]
    trial_ids_stopped = set()
    for _ in range(count):
      request = pythia.EarlyStopRequest(
          study_descriptor=self.policy_supporter.study_descriptor(),
          trial_ids=request_trial_ids,
      )
      early_stop_decisions = self.policy.early_stop(request)
      self.assertContainsSubset(
          request_trial_ids,
          [decision.id for decision in early_stop_decisions.decisions],
      )

      for decision in early_stop_decisions.decisions:
        if decision.should_stop:
          # Stop trials that need to be stopped.
          self.policy_supporter.trials[decision.id - 1].complete(
              pyvizier.Measurement()
          )
          trial_ids_stopped.add(decision.id)
    self.assertEqual(trial_ids_stopped, set(range(1, count + 1)))


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import itertools
from matplotlib.pylab import plt
import numpy as np
from vizier import benchmarks as vzb
from vizier._src.algorithms.designers import grid
from vizier._src.algorithms.designers import random
from vizier._src.benchmarks.analyzers import plot_utils
from vizier._src.benchmarks.analyzers import state_analyzer
from vizier.benchmarks import experimenters
from absl.testing import absltest


class PlotUtilsTest(absltest.TestCase):

  def test_plot_median_convergence(self):
    _, ax = plt.subplots(1, 1, figsize=(12, 8))
    plot_utils.plot_median_convergence(
        ax,
        curves=np.asarray([[1, 1, 2, 3, 4], [1, 1, 1, 2, float('nan')]]),
        percentiles=((40, 60), (30, 70)),
        alphas=(0.2, 0.4),
        xs=np.arange(1, 6),
        color='r',
    )

  def test_plot_mean_convergence(self):
    _, ax = plt.subplots(1, 1, figsize=(12, 8))
    plot_utils.plot_mean_convergence(
        ax,
        curves=np.asarray([[1, 1, 2, 3, 4], [1, 1, 1, 2, float('nan')]]),
        xs=np.arange(1, 6),
        color='r',
    )

  def test_plot_median_convergence_omit_optional_args(self):
    _, ax = plt.subplots(1, 1, figsize=(12, 8))
    plot_utils.plot_median_convergence(
        ax, curves=np.asarray([[1, 1, 2, 3, 4], [1, 1, 1, 2, float('nan')]])
    )

  def test_plot_records(self):
    function_names = ['Sphere', 'Discus']
    dimensions = [4, 8]
    product_list = list(itertools.product(function_names, dimensions))

    experimenter_factories = []
    for product in product_list:
      experimenter_factory = experimenters.BBOBExperimenterFactory(
          name=product[0], dim=product[1]
      )
      experimenter_factories.append(experimenter_factory)

    num_repeats = 5
    num_iterations = 20
    runner = vzb.BenchmarkRunner(
        benchmark_subroutines=[
            vzb.GenerateSuggestions(),
            vzb.EvaluateActiveTrials(),
        ],
        num_repeats=num_iterations,
    )
    algorithms = {
        'grid': grid.GridSearchDesigner.from_problem,
        'random': random.RandomDesigner.from_problem,
    }

    records = []
    for experimenter_factory in experimenter_factories:
      for algo_name, algo_factory in algorithms.items():
        benchmark_state_factory = vzb.ExperimenterDesignerBenchmarkStateFactory(
            experimenter_factory=experimenter_factory,
            designer_factory=algo_factory,
        )
        states = []
        for _ in range(num_repeats):
          benchmark_state = benchmark_state_factory()
          runner.run(benchmark_state)
          states.append(benchmark_state)
        record = state_analyzer.BenchmarkStateAnalyzer.to_record(
            algorithm=algo_name,
            experimenter_factory=experimenter_factory,
            states=states,
        )
        records.append(record)
    plot_utils.plot_from_records(records)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for lbfgsb_optimizer."""

import jax.numpy as jnp
import numpy as np
from vizier import pyvizier as vz
from vizier._src.algorithms.optimizers import lbfgsb_optimizer as lo
from vizier._src.algorithms.optimizers import vectorized_base as vb
from vizier.pyvizier import converters

from absl.testing import absltest
from absl.testing import parameterized


class LBFGSBOptimizer(parameterized.TestCase):

  def test_optimize_candidates_len(self):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('f1', 0.0, 10.0)
    problem.search_space.root.add_float_param('f2', 0.0, 10.0)
    problem.search_space.root.add_float_param('f3', 0.0, 10.0)
    converter = converters.TrialToModelInputConverter.from_problem(problem)
    score_fn = lambda x, _: jnp.sum(x.continuous.padded_array, axis=-1)
    optimizer = lo.LBFGSBOptimizerFactory(random_restarts=10, maxiter=20)(
        converter
    )
    res = optimizer(score_fn=score_fn)
    self.assertLen(res.rewards, 1)

  def test_best_candidates_count_is_1(self):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('f1', 0.0, 1.0)
    problem.search_space.root.add_float_param('f2', 0.0, 1.0)
    converter = converters.TrialToModelInputConverter.from_problem(problem)
    optimizer = lo.LBFGSBOptimizerFactory(random_restarts=10, maxiter=20)(
        converter
    )
    score_fn = lambda x, _: -jnp.sum(  # pylint: disable=g-long-lambda
        jnp.square(x.continuous.padded_array - 0.52), axis=-1
    )
    results = optimizer(score_fn=score_fn)
    candidates = vb.best_candidates_to_trials(results, converter)
    # check the best candidate
    self.assertLessEqual(
        np.abs(candidates[0].parameters['f1'].value - 0.52), 1e-6
    )
    self.assertLessEqual(
        np.abs(candidates[0].parameters['f2'].value - 0.52), 1e-6
    )
    self.assertIsNotNone(candidates[0].final_measurement)
    if candidates[0].final_measurement:
      self.assertLessEqual(
          np.abs(candidates[0].final_measurement.metrics['acquisition'].value),
          1e-6,
      )

  def test_batch_candidates(self):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('f1', 0.0, 1.0)
    problem.search_space.root.add_float_param('f2', 0.0, 1.0)
    problem.search_space.root.add_float_param('f3', 0.0, 1.0)
    converter = converters.TrialToModelInputConverter.from_problem(problem)
    # Minimize sum over all features.
    score_fn = lambda x, _: -jnp.sum(  # pylint: disable=g-long-lambda
        jnp.square(x.continuous.padded_array - 0.52),
        axis=[-1, -2],
    )
    optimizer = lo.LBFGSBOptimizerFactory(random_restarts=10, maxiter=20)(
        converter
    )
    best_results = optimizer(score_fn=score_fn, count=1, n_parallel=3)
    best_candidates = vb.best_candidates_to_trials(best_results, converter)
    self.assertLen(best_candidates, 3)
    # check the best candidates
    for b in best_candidates:
      self.assertLessEqual(np.abs(b.parameters['f1'].value - 0.52), 1e-6)
      self.assertLessEqual(np.abs(b.parameters['f2'].value - 0.52), 1e-6)
      self.assertIsNotNone(b.final_measurement)
      if b.final_measurement:
        self.assertLessEqual(
            np.abs(b.final_measurement.metrics['acquisition'].value), 1e-6
        )


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Experimenter that normalizes the range of each metric."""
import copy
from typing import Dict, Sequence

import numpy as np
from vizier import pyvizier
from vizier._src.benchmarks.experimenters import experimenter
from vizier.pyvizier import converters


# TODO: Improve normalization by using optimization.
class NormalizingExperimenter(experimenter.Experimenter):
  """Normalizes an Experimenter output via dividing by L1 norm at grid points."""

  def __init__(
      self,
      exptr: experimenter.Experimenter,
      num_normalization_samples: int = 100,
  ):
    """Normalizing experimenter uses a grid to estimate a normalization constant.

    Args:
      exptr: Experimenter to be normalized.
      num_normalization_samples: Number of samples to determine normalization.
    """
    self._exptr = exptr
    self._problem_statement = exptr.problem_statement()

    min_vals: list[float] = []
    max_vals: list[float] = []
    for pc in self._problem_statement.search_space.parameters:
      if pc.type != pyvizier.ParameterType.DOUBLE:
        raise ValueError(f'Only DOUBLE parameters can be normalized: {pc}.')
      min_value, max_value = pc.bounds
      min_vals.append(min_value)
      max_vals.append(max_value)

    grid = np.linspace(
        np.asarray(min_vals), np.asarray(max_vals), num_normalization_samples
    )
    converter = converters.TrialToArrayConverter.from_study_config(
        study_config=self._problem_statement, scale=False
    )
    grid_parameters = converter.to_parameters(grid)
    metrics: Dict[str, list[float]] = {}
    for parameters in grid_parameters:
      trial = pyvizier.Trial(parameters=parameters)
      exptr.evaluate([trial])
      measurement = trial.final_measurement
      for name, metric in (measurement.metrics if measurement else {}).items():
        if name in metrics:
          metrics[name].append(metric.value)
        else:
          metrics[name] = [metric.value]

    self._normalizations: Dict[str, float] = {}
    for name, grid_values in metrics.items():
      normalization = np.mean(np.absolute(np.array(grid_values)))
      if normalization == 0:
        raise ValueError(
            f'Cannot normalize {name} due to nonpositive L1 norm'
            f' with grid values {grid_values}'
        )
      self._normalizations[name] = normalization

  def problem_statement(self) -> pyvizier.ProblemStatement:
    return copy.deepcopy(self._problem_statement)

  def evaluate(self, suggestions: Sequence[pyvizier.Trial]):
    self._exptr.evaluate(suggestions)
    for suggestion in suggestions:
      if suggestion.final_measurement is None:
        continue
      normalized_metrics: Dict[str, pyvizier.Metric] = {}
      for name, metric in suggestion.final_measurement.metrics.items():
        normalized_metrics[name] = pyvizier.Metric(
            value=metric.value / self._normalizations[name]
        )
      suggestion.final_measurement.metrics = normalized_metrics

  def __repr__(self):
    return (
        f'NormalizingExperimenter with normalizations {self._normalizations} on'
        f' {self._exptr}'
    )



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for bocs."""
from vizier._src.algorithms.designers import harmonica
from vizier._src.algorithms.testing import test_runners
from vizier._src.benchmarks.experimenters import combo_experimenter

from absl.testing import absltest


class HarmonicaTest(absltest.TestCase):

  def test_make_suggestions(self):
    experimenter = combo_experimenter.IsingExperimenter(lamda=0.01)
    designer = harmonica.HarmonicaDesigner(
        experimenter.problem_statement(), num_init_samples=1)

    num_trials = 10
    trials = test_runners.run_with_random_metrics(
        designer,
        experimenter.problem_statement(),
        iters=num_trials,
        batch_size=1,
        verbose=1,
        validate_parameters=True)
    self.assertLen(trials, num_trials)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Policy Factory API."""
from typing import Protocol

from vizier import pyvizier as vz
from vizier._src.pythia import policy
from vizier._src.pythia import policy_supporter as policy_supporter_lib


class PolicyFactory(Protocol):
  """Protocol (PEP-544) for a Policy Factory."""

  def __call__(
      self,
      problem_statement: vz.ProblemStatement,
      algorithm: str,
      policy_supporter: policy_supporter_lib.PolicySupporter,
      study_name: str,
  ) -> policy.Policy:
    """Creates a Pythia Policy."""



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Scheduled designer for budget-aware optimization.
"""

import abc
import logging
from typing import Protocol, Sequence
import attrs
import numpy as np
from vizier import algorithms as vza
from vizier import pyvizier as vz
from vizier._src.pyvizier.shared import common
from vizier.interfaces import serializable

ParamsValues = dict[str, float]


@attrs.define
class ScheduledParam(abc.ABC):
  """Represents a scheduled parameter.

  The scheduling starts from `init_value` (the first step) and ends at
  `final_value` (last step).

  The `value` method returns the value of the scheduled param in the current
  step, accounting for the total number of steps in the scheduling process.

  Note that some types of scheduling (e.g. random, cyclic) are not supported.
  """

  init_value: float
  final_value: float

  def __attrs_post_init__(self):
    self._validate()

  def _validate(self) -> None:
    """Validate the scheduling param initial and final values."""
    total_steps = 10  # arbitrarily choose the total steps for validation.
    tol = 1e-5

    if abs(self.value(total_steps, 0) - self.init_value) > tol:
      raise ValueError("Initial value mismatch (step 0).")

    if abs(self.value(total_steps, total_steps - 1) - self.final_value) > tol:
      raise ValueError("Final value mismatch (last step).")

  @abc.abstractmethod
  def value(self, total_steps: int, step: int) -> float:
    """Compute the scheduling param value in the current step.

    Args:
      total_steps: The total number of steps for which the scheduling applied to
        (e.g. total number of study trials).
      step: The current step in the scheduling process (e.g. current study
        trial).
    """


@attrs.define
class LinearScheduledParam(ScheduledParam):
  """Linear scheduler."""

  def value(self, total_steps: int, step: int) -> float:
    slope = (self.final_value - self.init_value) / (total_steps - 1)
    return self.init_value + slope * step


@attrs.define
class ExponentialScheduledParam(ScheduledParam):
  """Exponential scheduler."""

  rate: float

  def value(self, total_steps: int, step: int) -> float:
    alpha = 1 / self.rate
    beta = -np.log(self.final_value / self.init_value) / (
        (total_steps - 1) ** alpha
    )
    return self.init_value * np.exp(-beta * step**alpha)


class DesignerStateUpdater(Protocol):
  """Update designer state based on params.

  Allows to efficiently modify the designer state without re-instantiating it.
  """

  def __call__(self, designer: vza.Designer, params: ParamsValues) -> None:
    """Update the designer state (inplace) based on parameter values.

    Note: The `params` dictionary does not need to directly correspond to
      `designer` attributes. The protocol implementations have the flexibility
      to determine how to use the `params`.

    Args:
      designer: The designer to update its state inplace.
      params: A dictionary of params to be used to update the designer state.
    """
    pass


@attrs.define
class ScheduledDesigner(vza.PartiallySerializableDesigner):
  """Scheduled designer."""

  _problem: vz.ProblemStatement = attrs.field(kw_only=False)
  _designer_factory: vza.DesignerFactory = attrs.field(kw_only=True)
  _designer_state_updater: DesignerStateUpdater = attrs.field(kw_only=True)
  _scheduled_params: dict[str, ScheduledParam] = attrs.field(kw_only=True)
  # The total number of study trials the designer is expected to generate. This
  # is used to determine the scheduling rate of change across the study.
  _expected_total_num_trials: int = attrs.field(
      kw_only=True, validator=attrs.validators.ge(0)
  )
  # ------------------------------------------------------------------
  # Internal attributes which should not be set by callers.
  # ------------------------------------------------------------------
  _designer: vza.Designer = attrs.field(init=False)
  _num_incorporated_suggested_trials: int = attrs.field(init=False, default=0)
  _metadata_ns: str = attrs.field(default="scheduled_designer", init=False)

  def __attrs_post_init__(self):
    self._designer = self._designer_factory(self._problem)
    self._update_designer_state()

  def dump(self) -> vz.Metadata:
    """Dumps the current state of the designer.

    Note that the state only contains the number of suggested trials. The other
    attributes are provided during designer instantiation.

    Returns:
      Metadata with the current designer's state.
    """
    metadata = vz.Metadata()
    metadata.ns(self._metadata_ns)["suggested_num_trials"] = str(
        self._num_incorporated_suggested_trials
    )
    return metadata

  def load(self, metadata: common.Metadata) -> None:
    """Loads the designer state from the metadata."""
    if (
        metadata.ns(self._metadata_ns).get("suggested_num_trials", default=None)
        is None
    ):
      # When the designer is called for the first time, or if the algorithm has
      # changed in the middle of the study.
      raise serializable.FatalDecodeError(
          "The metadata doesn't contain a state to be recovered."
      )
    try:
      self._num_incorporated_suggested_trials = int(
          metadata.ns(self._metadata_ns)["suggested_num_trials"]
      )
    except Exception as e:
      raise serializable.FatalDecodeError(
          "Couldn't load state ('suggested_num_trials') from metadata."
      ) from e

  @property
  def designer(self) -> vza.Designer:
    return self._designer

  @property
  def scheduled_params(self) -> dict[str, ScheduledParam]:
    return self._scheduled_params

  @property
  def num_incorporated_suggested_trials(self) -> int:
    """Returns the total number of suggested trials the designer generated (regardless of their current status)."""
    return self._num_incorporated_suggested_trials

  def update(
      self, completed: vza.CompletedTrials, all_active: vza.ActiveTrials
  ) -> None:
    """Update the underlying designer based on completed and pending trials."""
    self._designer.update(completed, all_active)
    self._validate_num_incorporated_suggested_trials(completed, all_active)

  def _validate_num_incorporated_suggested_trials(
      self, completed: vza.CompletedTrials, all_active: vza.ActiveTrials
  ) -> None:
    """Validate (and if needed update) the designer's suggested number of trials."""
    num_completed = len(completed.trials)
    num_active = len(all_active.trials)
    if self._num_incorporated_suggested_trials < num_completed + num_active:
      # If the designer is updated with more trials than what it observed, we
      # fast forward and update the state. Would happen if the designer is
      # updated with trials it didn't generate.
      logging.warning(
          "The scheduled designer's `num_incorporated_suggested_trials` (%s)"
          " doesn't match the actual number of trials in the study (%s).",
          self._num_incorporated_suggested_trials,
          num_completed + num_active,
      )
      self._num_incorporated_suggested_trials = num_completed + num_active

  def _update_designer_state(self) -> ParamsValues:
    """Efficiently update the underlying designer state (inplace).

    Returns:
      The current scheduled parameters values, which is used for logging.
    """
    params_values = {}
    for name, scheduled_param in self._scheduled_params.items():
      params_values[name] = scheduled_param.value(
          total_steps=self._expected_total_num_trials,
          step=self._num_incorporated_suggested_trials,
      )
    self._designer_state_updater(self._designer, params_values)
    logging.info("Updated designer state with params: %s", params_values)
    return params_values

  def suggest(self, count: int = 1) -> Sequence[vz.TrialSuggestion]:
    """Suggest trials."""
    # Compute the scheduled param values and update the designer state.
    params_values = self._update_designer_state()
    # Suggest 'count' trials in batch, using the same scheduled params values.
    suggest_trials = self._designer.suggest(count)
    self._num_incorporated_suggested_trials += len(suggest_trials)
    # Log the parameter values in the suggested trials for debugging.
    for trial in suggest_trials:
      metadata = trial.metadata.ns(self._metadata_ns).ns("devinfo")
      for name in self._scheduled_params.keys():
        metadata[name] = str(params_values[name])
    # Check that the maximum number of trials hasn't surpassed.
    if (
        self._num_incorporated_suggested_trials
        >= self._expected_total_num_trials
    ):
      logging.info(
          "Suggested trial count (%s) exceeded the configured maximum (%s).",
          self._num_incorporated_suggested_trials,
          self._expected_total_num_trials,
      )
    return suggest_trials



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for acquisitions."""

import jax
from jax import config
from jax import numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp
from vizier._src.algorithms.designers import scalarization
from vizier._src.algorithms.designers.gp import acquisitions
from vizier._src.jax import types
from absl.testing import absltest

tfd = tfp.distributions
tfpk = tfp.math.psd_kernels
tfpke = tfp.experimental.psd_kernels


def _make_test_model_data(labels_array, num_categorical=1):
  labels = types.PaddedArray.as_padded(labels_array)
  features = types.ModelInput(
      continuous=types.PaddedArray.as_padded(
          jnp.zeros((labels_array.shape[0], 3), dtype=jnp.float64)
      ),
      categorical=types.PaddedArray.as_padded(
          jnp.zeros(
              (labels_array.shape[0], num_categorical), dtype=types.INT_DTYPE
          ),
      ),
  )
  return types.ModelData(features=features, labels=labels)


class AcquisitionsTest(absltest.TestCase):

  def test_ucb(self):
    acq = acquisitions.UCB(coefficient=2.0)
    self.assertAlmostEqual(acq(tfd.Normal(0.1, 1)), 2.1)

  def test_lcb(self):
    acq = acquisitions.LCB(coefficient=2.0)
    self.assertAlmostEqual(acq(tfd.Normal(0.1, 1)), -1.9)

  def test_ei(self):
    labels = types.PaddedArray.as_padded(jnp.array([[0.2]]))
    best_labels = acquisitions.get_best_labels(labels)
    acq = acquisitions.EI(best_labels)
    self.assertAlmostEqual(
        acq(
            tfd.Normal(jnp.float64(0.1), 1),
        ),
        0.34635347,
    )

  def test_scalarized_ucb(self):
    labels = types.PaddedArray.as_padded(
        jnp.array([[0.2, 0.3], [0.01, 0.5], [0.5, 0.01]])
    )
    reference_point = acquisitions.get_worst_labels(labels)
    ucb = acquisitions.UCB(coefficient=0.1)
    scalarizer = scalarization.HyperVolumeScalarization(
        weights=jnp.array([0.1, 0.2]), reference_point=reference_point
    )

    acq = acquisitions.ScalarizeOverAcquisitions(ucb, scalarizer)
    self.assertAlmostEqual(
        acq(tfd.Normal([0.1, 0.2], [0.1, 0.1])), jnp.array([1.0]), delta=1e-2
    )

    # Tests that the scalarized acquisition is larger with max_scalarized.
    scalarized_labels = scalarizer(labels.unpad())
    max_scalarized = jnp.max(scalarized_labels, axis=-1)
    acq = acquisitions.ScalarizeOverAcquisitions(
        ucb, scalarizer, max_scalarized=max_scalarized
    )
    self.assertAlmostEqual(
        acq(tfd.Normal([0.1, 0.2], [0.1, 0.1])), jnp.array([2.10]), delta=1e-2
    )

  def test_ehvi_approximation(self):
    num_obj = 2
    num_scalarizations = 1000
    weights = jax.random.normal(
        jax.random.PRNGKey(0), shape=(num_scalarizations, num_obj)
    )
    weights = jnp.abs(weights)
    weights = weights / jnp.linalg.norm(weights, axis=1, keepdims=True)
    scalarizer = scalarization.HyperVolumeScalarization(weights)

    # Tests that the scalarizer gives the approximate hypervolume with mean
    # and uses constant rescaling of pi/4 for num_objs=2.
    hypervolume = acquisitions.ScalarizeOverAcquisitions(
        acquisitions.UCB(coefficient=0.0),
        scalarizer,
        reduction_fn=lambda x: jnp.mean(x, axis=0),
        max_scalarized=jnp.zeros(shape=(num_scalarizations,)),
    )
    # Expected hypervolume should be 2 * 1.5 = 3.0.
    np.testing.assert_allclose(
        hypervolume(tfd.Normal(jnp.array([2, 1.5]), jnp.ones(num_obj)))
        * (3.1415)
        / 4.0,
        jnp.array([3.0]),
        rtol=1e-1,
    )

  def test_ehvi_approximation_aq_over_scalar(self):
    num_obj = 2
    num_scalarizations = 1000
    weights = jax.random.normal(
        jax.random.PRNGKey(0), shape=(num_scalarizations, num_obj)
    )
    weights = jnp.abs(weights)
    weights = weights / jnp.linalg.norm(weights, axis=1, keepdims=True)
    scalarizer = scalarization.HyperVolumeScalarization(weights)

    # Tests that the scalarizer gives the approximate hypervolume with mean
    # and uses constant rescaling of pi/4 for num_objs=2.
    hypervolume = acquisitions.AcquisitionOverScalarized(
        acquisitions.UCB(coefficient=0.0), scalarizer
    )
    # Expected hypervolume should be 2 * 1.5 = 3.0.
    dist = tfd.Normal(jnp.array([2, 1.5]), jnp.ones(num_obj))
    np.testing.assert_allclose(
        hypervolume(dist, jax.random.PRNGKey(0)) * 3.1415 / 4.0,
        jnp.array([3.0]),
        rtol=1e-1,
    )

  def test_ehvi_mcmc(self):
    num_obj = 2
    num_scalarizations = 1000
    weights = jax.random.normal(
        jax.random.PRNGKey(0), shape=(num_scalarizations, num_obj)
    )
    weights = jnp.abs(weights)
    weights = weights / jnp.linalg.norm(weights, axis=1, keepdims=True)
    scalarizer = scalarization.HyperVolumeScalarization(weights)

    # Tests that the scalarizer gives the approximate hypervolume with mean
    # and uses constant rescaling of pi/4 for num_objs=2.
    hypervolume = acquisitions.ScalarizeOverAcquisitions(
        acquisitions.Sample(num_samples=100),
        scalarizer,
        reduction_fn=lambda x: jnp.mean(jax.nn.relu(x)),
    )
    # Expected hypervolume should be close to 2 * 1.5 = 3.0.
    stddev = 0.01 * jnp.ones(num_obj)
    np.testing.assert_allclose(
        hypervolume(tfd.Normal(jnp.array([2, 1.5]), stddev)) * (3.1415) / 4.0,
        jnp.array([3.0]),
        rtol=1e-1,
    )

  def test_pi(self):
    labels = types.PaddedArray.as_padded(jnp.array([[0.2]]))
    best_labels = acquisitions.get_best_labels(labels)
    acq = acquisitions.PI(best_labels)
    self.assertAlmostEqual(
        acq(
            tfd.Normal(jnp.float64(0.1), 1),
        ),
        0.46017216,
    )

  def test_mes(self):
    num_obs = 10
    num_pred = 6
    labels = np.random.normal(size=([num_obs, 1]))
    data = _make_test_model_data(labels, num_categorical=0)
    init_features = types.ModelInput(
        continuous=types.PaddedArray.as_padded(
            jnp.ones((num_pred, 3), dtype=jnp.float64)
        ),
        categorical=types.PaddedArray.as_padded(
            jnp.zeros((num_pred, 0), dtype=types.INT_DTYPE)
        ),
    )

    class _TestPredictive:

      def predict_with_aux(self, x):
        gp = tfd.GaussianProcess(
            kernel=tfpke.FeatureScaledWithCategorical(
                kernel=tfpk.ExponentiatedQuadratic(),
                scale_diag=tfpke.ContinuousAndCategoricalValues(
                    continuous=jnp.ones([3]), categorical=jnp.ones([0])
                ),
            ),
            index_points=tfpke.ContinuousAndCategoricalValues(
                continuous=data.features.continuous.padded_array,
                categorical=data.features.categorical.padded_array,
            ),
            observation_noise_variance=np.float64(1.0),
        )
        return (
            gp.posterior_predictive(
                observations=data.labels.padded_array[:, 0],
                predictive_index_points=tfpke.ContinuousAndCategoricalValues(
                    x.continuous.padded_array, x.categorical.padded_array
                ),
            ),
            {},
        )

    score_fn = acquisitions.MaxValueEntropySearch.scoring_fn_factory(
        data=data, predictive=_TestPredictive(), use_trust_region=True
    )
    score = score_fn.score(init_features, seed=jax.random.PRNGKey(0))
    self.assertEqual(score.shape, (num_pred,))

  def test_acq_pi_tr_good_point(self):
    data = _make_test_model_data((jnp.array([[100.0]])))
    acq = acquisitions.AcquisitionTrustRegion.default_ucb_pi(data=data)
    self.assertAlmostEqual(
        acq(tfd.Normal(jnp.float64(0.1), 1)),
        -1.0e12,
    )

  def test_acq_pi_tr_bad_point(self):
    data = _make_test_model_data((jnp.array([[-100.0]])))
    acq = acquisitions.AcquisitionTrustRegion.default_ucb_pi(data=data)
    self.assertAlmostEqual(
        acq(tfd.Normal(jnp.float64(0.1), 1)),
        1.9,
    )

  def test_acq_lcb_tr_bad_point(self):
    data = _make_test_model_data((jnp.array([[100.0]])))
    acq = acquisitions.AcquisitionTrustRegion.default_ucb_lcb(data=data)
    self.assertAlmostEqual(
        acq(tfd.Normal(jnp.float64(0.1), 1)),
        jnp.array([-1e12]),
        delta=2.0,
    )

  def test_acq_lcb_tr_good_point(self):
    data = _make_test_model_data((jnp.array([[-100.0]])))
    acq = acquisitions.AcquisitionTrustRegion.default_ucb_lcb(data=data)
    self.assertAlmostEqual(
        acq(tfd.Normal(jnp.float64(0.1), 1)),
        1.9,
    )

  def test_acq_lcb_delay_tr(self):
    data = _make_test_model_data((jnp.array([[100.0], [-100.0]])))
    acq = acquisitions.AcquisitionTrustRegion.default_ucb_lcb_delay_tr(
        data=data
    )
    self.assertAlmostEqual(
        acq(tfd.Normal(jnp.float64(0.1), 1)),
        jnp.array([1.9]),
    )

  def test_qei(self):
    best_labels = jnp.array([0.2])
    acq = acquisitions.QEI(num_samples=2000, best_labels=best_labels)
    batch_shape = [6]
    dist = tfd.Normal(loc=0.1 * jnp.ones(batch_shape), scale=1.0)
    qei = acq(dist, seed=jax.random.PRNGKey(0))
    # QEI reduces over the batch shape.
    self.assertEmpty(qei.shape)

    dist_single_point = tfd.Normal(jnp.array([0.1], dtype=jnp.float64), 1)
    qei_single_point = acq(dist_single_point, seed=jax.random.PRNGKey(0))
    # Parallel matches non-parallel for a single point.
    np.testing.assert_allclose(qei_single_point, 0.346, atol=1e-2)
    self.assertEmpty(qei_single_point.shape)

  def test_qpi(self):
    best_labels = jnp.array([0.2])
    acq = acquisitions.QPI(num_samples=5000, best_labels=best_labels)
    batch_shape = [6]
    dist = tfd.Normal(loc=0.1 * jnp.ones(batch_shape), scale=1.0)
    qpi = acq(dist, seed=jax.random.PRNGKey(0))
    # QPI reduces over the batch shape.
    self.assertEmpty(qpi.shape)

    dist_single_point = tfd.Normal(jnp.array([0.1], dtype=jnp.float64), 1)
    qpi_single_point = acq(dist_single_point, seed=jax.random.PRNGKey(0))
    # Parallel matches non-parallel for a single point.
    pi_single_point = acquisitions.PI(best_labels=best_labels)(
        dist_single_point
    )
    np.testing.assert_allclose(qpi_single_point, pi_single_point[0], atol=1e-2)
    self.assertEmpty(qpi_single_point.shape)

  def test_qucb_shape(self):
    acq = acquisitions.QUCB()
    batch_shape = [6]
    dist = tfd.Normal(loc=0.1 * jnp.ones(batch_shape), scale=1.0)
    qucb = acq(dist, seed=jax.random.PRNGKey(0))
    # QUCB reduces over the batch shape.
    self.assertEmpty(qucb.shape)

  def test_qucb_equals_ucb(self):
    # The QUCB coefficient should be multiplied by sqrt(pi/2) for equivalency
    # with the UCB coefficient (assuming a Gaussian distribution).
    acq_ucb = acquisitions.UCB(coefficient=0.5)
    acq_qucb = acquisitions.QUCB(
        num_samples=5000,
        coefficient=0.5 * np.sqrt(np.pi / 2.0),
    )
    dist_single_point = tfd.Normal(jnp.array([0.1], dtype=jnp.float64), 1)
    qucb_single_point = acq_qucb(dist_single_point, seed=jax.random.PRNGKey(1))
    ucb_single_point = acq_ucb(dist_single_point)
    # Parallel matches non-parallel for a single point.
    np.testing.assert_allclose(
        qucb_single_point, ucb_single_point[0], atol=2e-2
    )
    self.assertEmpty(qucb_single_point.shape)

  def test_multi_acquisition(self):
    best_labels = jnp.array([0.2])
    ucb = acquisitions.UCB()
    ei = acquisitions.EI(best_labels=best_labels)
    acq = acquisitions.MultiAcquisitionFunction({'ucb': ucb, 'ei': ei})
    dist = tfd.Normal(jnp.float64(0.1), 1)
    acq_val = acq(dist)
    ucb_val = ucb(dist)
    ei_val = ei(dist)
    np.testing.assert_allclose(acq_val, jnp.stack([ucb_val, ei_val]))


class TrustRegionTest(absltest.TestCase):

  def test_trust_region_small(self):
    trusted = types.ModelInput(
        continuous=types.PaddedArray.as_padded(
            np.array([[0.0, 0.0], [1.0, 1.0]]),
        ),
        categorical=types.PaddedArray.as_padded(np.array([[0, 0], [1, 1]])),
    )
    tr = acquisitions.TrustRegion(trusted=trusted)

    xs = types.ModelInput(
        continuous=types.PaddedArray.as_padded(
            np.array([
                [0.0, 0.3],
                [0.9, 0.8],
                [1.0, 1.0],
            ]),
        ),
        categorical=types.PaddedArray.as_padded(
            np.array([[0, 0], [1, 1], [0, 1]]),
        ),
    )
    np.testing.assert_allclose(
        tr.min_linf_distance(xs),
        np.array([0.3, 0.2, 0.0]),
    )
    self.assertAlmostEqual(tr.trust_radius, 0.224, places=3)

  def test_trust_region_bigger(self):
    xs_cont = np.vstack(
        [
            [0.0, 0.0],
            [1.0, 1.0],
        ]
        * 10
    )
    xs = types.ModelInput(
        continuous=types.PaddedArray.as_padded(xs_cont),
        categorical=types.PaddedArray.as_padded(
            np.ones(xs_cont.shape, dtype=types.INT_DTYPE),
        ),
    )
    tr = acquisitions.TrustRegion(trusted=xs)

    xs_cont_test = np.array([[0.0, 0.3], [0.9, 0.8], [1.0, 1.0]])
    xs_test = types.ModelInput(
        continuous=types.PaddedArray.as_padded(xs_cont_test),
        categorical=types.PaddedArray.as_padded(
            np.ones(xs_cont_test.shape, dtype=types.INT_DTYPE),
        ),
    )
    np.testing.assert_allclose(
        tr.min_linf_distance(xs_test),
        np.array([0.3, 0.2, 0.0]),
    )
    self.assertAlmostEqual(tr.trust_radius, 0.44, places=3)

  def test_trust_region_padded_small(self):
    # Test that padding still retrieves the same distance computations as
    # `test_trust_region_small`.
    xs_cont = np.array([
        [0.0, 0.0],
        [1.0, 1.0],
    ])
    xs = types.ModelInput(
        continuous=types.PaddedArray.from_array(
            xs_cont, target_shape=(4, 6), fill_value=0.0
        ),
        categorical=types.PaddedArray.from_array(
            np.ones(xs_cont.shape, dtype=types.INT_DTYPE),
            target_shape=(4, 5),
            fill_value=0,
        ),
    )
    tr = acquisitions.TrustRegion(
        trusted=xs,
    )

    xs_cont_test = np.array([
        [[0.0, 0.3], [0.9, 0.8], [1.0, 1.0]],
        [[1.0, 1.0], [0.0, 0.3], [0.9, 0.8]],
    ])
    xs_test = types.ModelInput(
        continuous=types.PaddedArray.from_array(
            xs_cont_test, target_shape=(3, 3, 6), fill_value=-100.0
        ),
        categorical=types.PaddedArray.from_array(
            np.ones(xs_cont_test.shape, dtype=types.INT_DTYPE),
            target_shape=(3, 3, 5),
            fill_value=-100,
        ),
    )
    np.testing.assert_allclose(
        tr.min_linf_distance(xs_test),
        np.array([[0.3, 0.2, 0.0], [0.0, 0.3, 0.2], [0.0, 0.0, 0.0]]),
    )
    self.assertAlmostEqual(tr.trust_radius, 0.224, places=3)

  def test_trust_region_padded_all_categorical_multi_batch_dims(self):
    xs = types.ModelInput(
        continuous=types.PaddedArray.as_padded(jnp.array([])),
        categorical=types.PaddedArray.from_array(
            np.ones((2, 2), dtype=types.INT_DTYPE),
            target_shape=(4, 5),
            fill_value=0,
        ),
    )
    tr = acquisitions.TrustRegion(
        trusted=xs,
    )
    xs_test = types.ModelInput(
        continuous=types.PaddedArray.as_padded(jnp.array([])),
        categorical=types.PaddedArray.from_array(
            np.ones((2, 3, 2), dtype=types.INT_DTYPE),
            target_shape=(3, 3, 5),
            fill_value=-100,
        ),
    )
    np.testing.assert_allclose(
        tr.min_linf_distance(xs_test),
        np.ones((3, 3)) * -np.inf,
    )


if __name__ == '__main__':
  config.update('jax_enable_x64', True)
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for grid."""
import functools

from vizier import pythia
from vizier import pyvizier
from vizier._src.algorithms.designers import grid
from vizier._src.algorithms.policies import designer_policy
from vizier._src.algorithms.testing import test_runners
from vizier.interfaces import serializable

from absl.testing import absltest
from absl.testing import parameterized


class GridSearchTest(parameterized.TestCase):
  # TODO: Add conditional test case.

  def setUp(self):
    """Setups up search space."""
    self.search_space = pyvizier.SearchSpace()
    self.search_space.root.add_float_param(
        'double', min_value=-1.0, max_value=1.0
    )
    self.search_space.root.add_float_param(
        'double_logscaled',
        min_value=0.001,
        max_value=1.0,
        scale_type=pyvizier.ScaleType.LOG,
    )
    self.search_space.root.add_float_param(
        'double_same_bounds', min_value=30.0, max_value=30.0
    )
    self.search_space.root.add_categorical_param(
        name='categorical', feasible_values=['a', 'b', 'c']
    )
    self.search_space.root.add_discrete_param(
        name='discrete', feasible_values=[0.1, 0.3, 0.5]
    )
    self.search_space.root.add_int_param(name='int', min_value=1, max_value=5)
    self.search_space.root.add_int_param(
        name='int_same_bounds', min_value=150, max_value=150
    )
    self._designer = grid.GridSearchDesigner(self.search_space)
    self.search_space_size = (
        self._designer._double_grid_resolution
        * self._designer._double_grid_resolution
        * 3
        * 3
        * 5
    )
    super().setUp()

  def test_make_grid_values(self):
    grid_values = self._designer._grid_values
    self.assertLen(
        grid_values['double'], self._designer._double_grid_resolution
    )
    double_list = [p.value for p in grid_values['double']]
    for d in double_list:
      self.assertBetween(d, -1.0, 1.0)
    self.assertContainsSubset([-1.0, 1.0], double_list)
    self.assertLen(grid_values['categorical'], 3)
    self.assertLen(grid_values['discrete'], 3)
    self.assertLen(grid_values['int'], 5)

  def test_load_failure(self):
    problem = pyvizier.ProblemStatement(self.search_space)
    test_runners.RandomMetricsRunner(
        problem, iters=20, validate_parameters=True
    ).run_designer(self._designer)

    with self.assertRaises(serializable.HarmlessDecodeError):
      self._designer.load(pyvizier.Metadata())

    # After the load failure, the grid designer is still in a valid state
    # and can be used to generate suggestions.
    test_runners.RandomMetricsRunner(
        problem, iters=10, validate_parameters=True
    ).run_designer(self._designer)

  @parameterized.parameters((None,), (0,), (1,), (2,))
  def test_make_suggestions(self, shuffle_seed):
    """Tests designer suggestion generation."""
    designer = grid.GridSearchDesigner(self.search_space, shuffle_seed)
    suggestions = designer.suggest(self.search_space_size)
    self.assertLen(suggestions, self.search_space_size)
    print(suggestions)
    for suggestion in suggestions:
      self.assertLen(suggestion.parameters, len(self.search_space.parameters))

    # Make sure we covered entire search space.
    distinct_suggestions = set(
        [
            tuple(suggestion.parameters.as_dict().values())
            for suggestion in suggestions
        ]
    )
    self.assertLen(distinct_suggestions, self.search_space_size)

  @parameterized.parameters((None,), (0,), (1,), (2,))
  def test_policy_wrapping(self, shuffle_seed):
    problem = pyvizier.ProblemStatement()
    problem.search_space = self.search_space
    policy_supporter = pythia.InRamPolicySupporter(problem)
    designer_factory = functools.partial(
        grid.GridSearchDesigner.from_problem, seed=shuffle_seed
    )
    policy = designer_policy.PartiallySerializableDesignerPolicy(
        problem, policy_supporter, designer_factory
    )

    # Make sure we covered entire search space.
    all_suggestions = []
    for _ in range(self.search_space_size):
      request = pythia.SuggestRequest(
          study_descriptor=policy_supporter.study_descriptor(), count=1
      )
      decisions = policy.suggest(request)
      all_suggestions.extend(decisions.suggestions)

    distinct_suggestions = set(
        [
            tuple(suggestion.parameters.as_dict().values())
            for suggestion in all_suggestions
        ]
    )
    self.assertLen(distinct_suggestions, self.search_space_size)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Input warping.

Input warping has shown to help optimize non-stationary functions:
http://proceedings.mlr.press/v32/snoek14.pdf.

Note that the dth output dimension is a function of the dth input dimension.

Example
-------
search_space = ...
problem = vz.ProblemStatement(search_space)
converter = converters.TrialToArrayConverter.from_study_config(problem)
input_warper = input_warping.KumaraswamyInputWarpingConverter(
    converter, a=0.1, b=0.8
)
trials = ...
features = input_warper.to_features(trials)
"""

from typing import Sequence

import attr
import numpy as np
from vizier import pyvizier as vz
from vizier.pyvizier.converters import core


def kumaraswamy_cdf(x: np.ndarray, a: float, b: float) -> np.ndarray:
  """Compute the Kumaraswamy CDF.

  Arguments:
    x: values in [0,1]. shape: (num_samples, num_features)
    a: positive value.
    b: positive value.

  Returns:
    The CDF(x). shape: (num_samples, num_cdfs).
  """
  return 1 - (1 - x**a) ** b


def kumaraswamy_inv_cdf(f: np.ndarray, a: float, b: float) -> np.ndarray:
  """Compute the inverse of the Kumaraswamy CDF.

  Arguments:
    f: values in [0,1]. shape: (num_samples, num_cdfs)
    a: positive value.
    b: positive value.

  Returns:
    The Inv_CDF(x). shape: (num_samples, num_features).
  """
  return (1 - (1 - f) ** (1 / b)) ** (1 / a)


@attr.define
class KumaraswamyInputWarpingConverter:
  """Input Warping based on Kumaraswamy distribution.

  Reference: https://en.wikipedia.org/wiki/Kumaraswamy_distribution
  """

  converter: core.TrialToArrayConverter
  a: float = attr.field(validator=attr.validators.instance_of(float))
  b: float = attr.field(validator=attr.validators.instance_of(float))

  def __attrs_post_init__(self):
    if self.a <= 0:
      raise ValueError(f"Attribute 'a' has to be positive, received {self.a}")
    if self.b <= 0:
      raise ValueError(f"Attribute 'b' has to be positive, received {self.b}")

  def to_features(self, trials) -> np.ndarray:
    features = self.converter.to_features(trials)
    return kumaraswamy_cdf(features, self.a, self.b)

  def to_xy(self, trials) -> tuple[np.ndarray, np.ndarray]:
    return self.to_features(trials), self.converter.to_labels(trials)

  def to_parameters(self, features: np.ndarray) -> Sequence[vz.ParameterDict]:
    """Convert to warput features into parameters."""
    return self.converter.to_parameters(
        kumaraswamy_inv_cdf(features, self.a, self.b)
    )



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Multi-objective synthetic functions created by Deb et al.

Forked from
https://github.com/pytorch/botorch/blob/main/botorch/test_functions/multi_objective.py.
"""

from typing import Callable, Sequence
import numpy as np
from vizier import pyvizier as vz
from vizier._src.benchmarks.experimenters import experimenter
from vizier.pyvizier import converters


# TODO: Subclass MultiObjectiveNumpyExperimenter.
class DHExperimenter(experimenter.Experimenter):
  """Multiobjective problem with two objectives (f0, f1) and d-dimensions.

  The functions are defined as:
    f0(x) = x0
    f1(x) = h(x) + g(x) * s(x) for DH1 and DH2
    f1(x) = h(x) * (g(x) + s(x)) for DH3 and DH4

  Reference:
    K. Deb and H. Gupta. Searching for Robust Pareto-Optimal Solutions in
    Multi-objective Optimization. Evolutionary Multi-Criterion Optimization,
    Springer-Berlin, pp. 150-164, 2005.
  """

  def __init__(
      self,
      h_fn: Callable[[np.ndarray], float],
      g_fn: Callable[[np.ndarray], float],
      s_fn: Callable[[np.ndarray], float],
      f1_fn: Callable[[float, float, float], float],  # Uses h, g, s.
      bounds: Sequence[tuple[float, float]],
  ):
    self._h_fn = h_fn
    self._g_fn = g_fn
    self._s_fn = s_fn
    self._f1_fn = f1_fn
    self._bounds = bounds

    self._converter = converters.TrialToArrayConverter.from_study_config(
        study_config=self.problem_statement(),
        scale=False,
        flip_sign_for_minimization_metrics=False,
    )

  def problem_statement(self) -> vz.ProblemStatement:
    problem = vz.ProblemStatement()
    problem.metric_information.append(
        vz.MetricInformation(name="f0", goal=vz.ObjectiveMetricGoal.MINIMIZE)
    )
    problem.metric_information.append(
        vz.MetricInformation(name="f1", goal=vz.ObjectiveMetricGoal.MINIMIZE)
    )
    for i, (min_val, max_val) in enumerate(self._bounds):
      problem.search_space.root.add_float_param(f"x{i}", min_val, max_val)

    return problem

  def evaluate(self, suggestions: Sequence[vz.Trial]) -> None:
    features = self._converter.to_features(suggestions)
    for i, suggestion in enumerate(suggestions):
      feat = features[i]
      f0 = feat[0]
      f1 = self._f1_fn(self._h_fn(feat), self._g_fn(feat), self._s_fn(feat))
      suggestion.complete(vz.Measurement(metrics={"f0": f0, "f1": f1}))

  @classmethod
  def DH1(cls, num_dimensions: int) -> "DHExperimenter":
    if num_dimensions < 2:
      raise ValueError(f"num_dimensions must be >= 2, got {num_dimensions}.")
    h_fn = lambda x: 1 - x[0] ** 2
    g_fn = lambda x: np.sum(
        10 + x[1:] ** 2 - 10 * np.cos(4 * np.pi * x[1:]),
    )
    s_fn = lambda x: 1 / (0.2 + x[0]) + x[0] ** 2
    f1_fn = lambda h, g, s: h + g * s
    bounds = [(0, 1)] + [(-1, 1) for _ in range(num_dimensions - 1)]
    return cls(h_fn, g_fn, s_fn, f1_fn, bounds)

  @classmethod
  def DH2(cls, num_dimensions: int) -> "DHExperimenter":
    """Same as DH1 but s_fn has a different (10.0) constant."""
    if num_dimensions < 2:
      raise ValueError(f"num_dimensions must be >= 2, got {num_dimensions}.")
    h_fn = lambda x: 1 - x[0] ** 2
    g_fn = lambda x: np.sum(
        10 + x[1:] ** 2 - 10 * np.cos(4 * np.pi * x[1:]),
    )
    s_fn = lambda x: 1 / (0.2 + x[0]) + 10.0 * x[0] ** 2
    f1_fn = lambda h, g, s: h + g * s
    bounds = [(0, 1)] + [(-1, 1) for _ in range(num_dimensions - 1)]
    return cls(h_fn, g_fn, s_fn, f1_fn, bounds)

  @classmethod
  def DH3(cls, num_dimensions: int) -> "DHExperimenter":
    if num_dimensions < 3:
      raise ValueError(f"num_dimensions must be >= 3, got {num_dimensions}.")
    exp_arg1 = lambda x: -(((x[1] - 0.35) / 0.25) ** 2)
    exp_arg2 = lambda x: -(((x[1] - 0.85) / 0.03) ** 2)
    h_fn = lambda x: 2 - 0.8 * np.exp(exp_arg1(x)) - np.exp(exp_arg2(x))

    g_fn = lambda x: 50 * np.sum(x[2:] ** 2)
    s_fn = lambda x: 1 - np.sqrt(x[0])
    f1_fn = lambda h, g, s: h * (g + s)
    bounds = [(0, 1), (0, 1)] + [(-1, 1) for _ in range(num_dimensions - 2)]
    return cls(h_fn, g_fn, s_fn, f1_fn, bounds)

  @classmethod
  def DH4(cls, num_dimensions: int) -> "DHExperimenter":
    """Similar to DH3 but different h_fn."""
    if num_dimensions < 3:
      raise ValueError(f"num_dimensions must be >= 3, got {num_dimensions}.")
    exp_arg1 = lambda x: -(((np.sum(x[:2]) - 0.35) / 0.25) ** 2)
    exp_arg2 = lambda x: -(((np.sum(x[:2]) - 0.85) / 0.03) ** 2)
    h_fn = lambda x: 2 - x[0] - 0.8 * np.exp(exp_arg1(x)) - np.exp(exp_arg2(x))

    g_fn = lambda x: 50 * np.sum(x[2:] ** 2)
    s_fn = lambda x: 1 - np.sqrt(x[0])
    f1_fn = lambda h, g, s: h * (g + s)
    bounds = [(0, 1), (0, 1)] + [(-1, 1) for _ in range(num_dimensions - 2)]
    return cls(h_fn, g_fn, s_fn, f1_fn, bounds)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Scalarizing Designer for MOO via reducing to a scalarized objective."""
import copy
import random
from typing import Optional, Sequence

from jax import numpy as jnp
from jax import random as jax_random
from vizier import algorithms as vza
from vizier import pyvizier as vz
from vizier._src.algorithms.designers import scalarization
from vizier._src.algorithms.ensemble import ensemble_designer


class ScalarizingDesigner(vza.Designer):
  """Multiobjective Scalarizing Designer.

  This designer applies multiobjective optimization by reducing to
  single objective optimization via scalarization.
  """

  def __init__(
      self,
      problem_statement: vz.ProblemStatement,
      designer_factory: vza.DesignerFactory[vza.Designer],
      scalarizer: scalarization.Scalarization,
      *,
      seed: Optional[int] = None,
  ):
    """Init.

    Args:
      problem_statement: Must be a mulitobjective search space.
      designer_factory: Factory to create the single-objective designer.
      scalarizer: Scalarization to be applied to objective metrics.
      seed: Any valid seed for factory.
    """
    self._scalarizer = scalarizer
    self._objectives = problem_statement.metric_information.of_type(
        vz.MetricType.OBJECTIVE
    )
    if len(self._objectives) <= 1:
      raise ValueError(f'Problem should be multi-objective {self._objectives}')

    # Create a single-objective designer.
    self._scalarized_metric_name = 'scalarized'
    single_objective_metric = problem_statement.metric_information.exclude_type(
        vz.MetricType.OBJECTIVE
    )
    single_objective_metric.append(
        vz.MetricInformation(
            name=self._scalarized_metric_name,
            goal=vz.ObjectiveMetricGoal.MAXIMIZE,
        )
    )
    self._problem_statement = copy.deepcopy(problem_statement)
    self._problem_statement.metric_information = single_objective_metric
    self._designer = designer_factory(self._problem_statement, seed=seed)

  def update(
      self, completed: vza.CompletedTrials, all_active: vza.ActiveTrials
  ) -> None:
    for trial in completed.trials:
      objectives = [
          trial.final_measurement_or_die.metrics.get_value(
              config.name, default=jnp.nan
          )
          for config in self._objectives
      ]
      # Simply append the scalarized value.
      trial.final_measurement_or_die.metrics[self._scalarized_metric_name] = (
          self._scalarizer(jnp.array(objectives))
      )

    self._designer.update(completed, all_active)

  def suggest(
      self, count: Optional[int] = None
  ) -> Sequence[vz.TrialSuggestion]:
    """Make new suggestions.

    Args:
      count: Makes best effort to generate this many suggestions. If None,
        suggests as many as the algorithm wants.

    Returns:
      New suggestions.
    """
    return self._designer.suggest(count)


def create_gaussian_scalarizing_designer(
    problem_statement: vz.ProblemStatement,
    designer_factory: vza.DesignerFactory[vza.Designer],
    scalarization_factory: scalarization.ScalarizationFromWeights,
    num_ensemble: int,
    *,
    seed: Optional[int] = None,
) -> vza.Designer:
  """Factory for an Ensemble of Gaussian weighted scalarized Designers."""
  objectives = problem_statement.metric_information.of_type(
      vz.MetricType.OBJECTIVE
  )
  if len(objectives) <= 1:
    raise ValueError(
        'Problem should be multi-objective for applying '
        f'scalarization ensembling {objectives}'
    )

  key = jax_random.PRNGKey(seed or random.getrandbits(32))
  weights = abs(
      jax_random.normal(key=key, shape=(num_ensemble, len(objectives)))
  )
  weights /= jnp.linalg.norm(weights, axis=1)[..., jnp.newaxis]
  ensemble_dict = {}
  for weight in weights:
    ensemble_dict[f'scalarized_weight: {weight}'] = ScalarizingDesigner(
        problem_statement=problem_statement,
        designer_factory=designer_factory,
        scalarizer=scalarization_factory(weight),
        seed=seed,
    )
  return ensemble_designer.EnsembleDesigner(ensemble_dict)
